#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any
from unittest import mock
from unittest.mock import call

import pytest
from google.api_core.exceptions import NotFound
from google.api_core.gapic_v1.method import DEFAULT

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.operators.alloy_db import (
    AlloyDBBaseOperator,
    AlloyDBCreateBackupOperator,
    AlloyDBCreateClusterOperator,
    AlloyDBCreateInstanceOperator,
    AlloyDBCreateUserOperator,
    AlloyDBDeleteBackupOperator,
    AlloyDBDeleteClusterOperator,
    AlloyDBDeleteInstanceOperator,
    AlloyDBDeleteUserOperator,
    AlloyDBUpdateBackupOperator,
    AlloyDBUpdateClusterOperator,
    AlloyDBUpdateInstanceOperator,
    AlloyDBUpdateUserOperator,
    AlloyDBWriteBaseOperator,
)

TEST_TASK_ID = "test-task-id"
TEST_GCP_PROJECT = "test-project"
TEST_GCP_REGION = "global"
TEST_GCP_CONN_ID = "test_conn_id"
TEST_IMPERSONATION_CHAIN = "test_impersonation_chain"
TEST_RETRY = DEFAULT
TEST_TIMEOUT = None
TEST_METADATA = ()

TEST_REQUEST_ID = "test_request_id"
TEST_VALIDATE_ONLY = False

TEST_CLUSTER_ID = "test_cluster_id"
TEST_CLUSTER_NAME = f"projects/{TEST_GCP_PROJECT}/locations/{TEST_GCP_REGION}/clusters/{TEST_CLUSTER_ID}"
TEST_CLUSTER: dict[str, Any] = {}
TEST_IS_SECONDARY = False
TEST_UPDATE_MASK = None
TEST_ALLOW_MISSING = False
TEST_ETAG = "test-etag"
TEST_FORCE = False

TEST_INSTANCE_ID = "test_instance_id"
TEST_INSTANCE: dict[str, Any] = {}

TEST_USER_ID = "test_user"
TEST_USER: dict[str, Any] = {}

TEST_BACKUP_ID = "test_backup_id"
TEST_BACKUP: dict[str, Any] = {}

OPERATOR_MODULE_PATH = "airflow.providers.google.cloud.operators.alloy_db.{}"
ALLOY_DB_HOOK_PATH = OPERATOR_MODULE_PATH.format("AlloyDbHook")
BASE_WRITE_CLUSTER_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBWriteBaseOperator.{}")
CREATE_CLUSTER_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBCreateClusterOperator.{}")
UPDATE_CLUSTER_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBUpdateClusterOperator.{}")
DELETE_CLUSTER_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBDeleteClusterOperator.{}")

CREATE_INSTANCE_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBCreateInstanceOperator.{}")
UPDATE_INSTANCE_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBUpdateInstanceOperator.{}")
DELETE_INSTANCE_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBDeleteInstanceOperator.{}")

CREATE_USER_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBCreateUserOperator.{}")
UPDATE_USER_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBUpdateUserOperator.{}")
DELETE_USER_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBDeleteUserOperator.{}")

CREATE_BACKUP_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBCreateBackupOperator.{}")
UPDATE_BACKUP_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBUpdateBackupOperator.{}")
DELETE_BACKUP_OPERATOR_PATH = OPERATOR_MODULE_PATH.format("AlloyDBDeleteBackupOperator.{}")


class TestAlloyDBBaseOperator:
    def setup_method(self):
        self.operator = AlloyDBBaseOperator(
            task_id=TEST_TASK_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.project_id == TEST_GCP_PROJECT
        assert self.operator.location == TEST_GCP_REGION
        assert self.operator.gcp_conn_id == TEST_GCP_CONN_ID
        assert self.operator.impersonation_chain == TEST_IMPERSONATION_CHAIN
        assert self.operator.retry == TEST_RETRY
        assert self.operator.timeout == TEST_TIMEOUT
        assert self.operator.metadata == TEST_METADATA

    def test_template_fields(self):
        expected_template_fields = {"project_id", "location", "gcp_conn_id"}
        assert set(AlloyDBBaseOperator.template_fields) == expected_template_fields

    @mock.patch(ALLOY_DB_HOOK_PATH)
    def test_hook(self, mock_hook):
        expected_hook = mock_hook.return_value

        hook_1 = self.operator.hook
        hook_2 = self.operator.hook

        mock_hook.assert_called_once_with(
            gcp_conn_id=TEST_GCP_CONN_ID, impersonation_chain=TEST_IMPERSONATION_CHAIN
        )
        assert hook_1 == expected_hook
        assert hook_2 == expected_hook


class TestAlloyDBWriteBaseOperator:
    def setup_method(self):
        self.operator = AlloyDBWriteBaseOperator(
            task_id=TEST_TASK_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.request_id == TEST_REQUEST_ID
        assert self.operator.validate_request == TEST_VALIDATE_ONLY

    def test_template_fields(self):
        expected_template_fields = {"request_id", "validate_request"} | set(
            AlloyDBBaseOperator.template_fields
        )
        assert set(AlloyDBWriteBaseOperator.template_fields) == expected_template_fields

    @mock.patch(BASE_WRITE_CLUSTER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH)
    def test_get_operation_result(self, mock_hook, mock_log):
        mock_operation = mock.MagicMock()
        mock_wait_for_operation = mock_hook.return_value.wait_for_operation
        expected_result = mock_wait_for_operation.return_value

        result = self.operator.get_operation_result(mock_operation)

        assert result == expected_result
        assert not mock_log.called
        mock_wait_for_operation.assert_called_once_with(timeout=TEST_TIMEOUT, operation=mock_operation)

    @mock.patch(BASE_WRITE_CLUSTER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH)
    def test_get_operation_result_validate_result(self, mock_hook, mock_log):
        mock_operation = mock.MagicMock()
        mock_wait_for_operation = mock_hook.return_value.wait_for_operation
        self.operator.validate_request = True

        result = self.operator.get_operation_result(mock_operation)

        assert result is None
        assert not mock_log.info.called
        assert not mock_wait_for_operation.called


class TestAlloyDBCreateClusterOperator:
    def setup_method(self):
        self.operator = AlloyDBCreateClusterOperator(
            task_id=TEST_TASK_ID,
            cluster_id=TEST_CLUSTER_ID,
            cluster_configuration=TEST_CLUSTER,
            is_secondary=TEST_IS_SECONDARY,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.cluster_id == TEST_CLUSTER_ID
        assert self.operator.cluster_configuration == TEST_CLUSTER
        assert self.operator.is_secondary == TEST_IS_SECONDARY

    def test_template_fields(self):
        expected_template_fields = {"cluster_id", "is_secondary", "cluster_configuration"} | set(
            AlloyDBWriteBaseOperator.template_fields
        )
        assert set(AlloyDBCreateClusterOperator.template_fields) == expected_template_fields

    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_get_cluster_not_found(self, mock_hook, mock_log):
        mock_get_cluster = mock_hook.return_value.get_cluster
        mock_get_cluster.side_effect = NotFound("Not found")

        result = self.operator._get_cluster()

        mock_get_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
        )
        mock_log.info.assert_has_calls(
            [
                call("Checking if the cluster %s exists already...", TEST_CLUSTER_ID),
                call("The cluster %s does not exist yet.", TEST_CLUSTER_ID),
            ]
        )
        assert result is None

    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_get_cluster_exception(self, mock_hook, mock_log):
        mock_get_cluster = mock_hook.return_value.get_cluster
        mock_get_cluster.side_effect = Exception()

        with pytest.raises(AirflowException):
            self.operator._get_cluster()

        mock_get_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
        )
        mock_log.info.assert_called_once_with("Checking if the cluster %s exists already...", TEST_CLUSTER_ID)

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Cluster.to_dict"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_get_cluster(self, mock_hook, mock_log, mock_to_dict):
        mock_get_cluster = mock_hook.return_value.get_cluster
        mock_cluster = mock_get_cluster.return_value
        expected_result = mock_to_dict.return_value

        result = self.operator._get_cluster()

        mock_get_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
        )
        mock_log.info.assert_has_calls(
            [
                call("Checking if the cluster %s exists already...", TEST_CLUSTER_ID),
                call("AlloyDB cluster %s already exists.", TEST_CLUSTER_ID),
            ]
        )
        mock_to_dict.assert_called_once_with(mock_cluster)
        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Cluster.to_dict"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("_get_cluster"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute(self, mock_hook, mock_log, mock_get_operation_result, mock_get_cluster, mock_to_dict):
        mock_get_cluster.return_value = None
        mock_create_cluster = mock_hook.return_value.create_cluster
        mock_create_secondary_cluster = mock_hook.return_value.create_secondary_cluster
        mock_operation = mock_create_cluster.return_value
        mock_operation_result = mock_get_operation_result.return_value

        expected_result = mock_to_dict.return_value
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Creating an AlloyDB cluster.")
        mock_get_cluster.assert_called_once()
        mock_create_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            cluster=TEST_CLUSTER,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_create_secondary_cluster.called
        mock_to_dict.assert_called_once_with(mock_operation_result)
        mock_get_operation_result.assert_called_once_with(mock_operation)

        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Cluster.to_dict"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("_get_cluster"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_is_secondary(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_cluster,
        mock_to_dict,
    ):
        mock_get_cluster.return_value = None
        mock_create_cluster = mock_hook.return_value.create_cluster
        mock_create_secondary_cluster = mock_hook.return_value.create_secondary_cluster
        mock_operation = mock_create_secondary_cluster.return_value
        mock_operation_result = mock_get_operation_result.return_value

        expected_result = mock_to_dict.return_value
        mock_context = mock.MagicMock()
        self.operator.is_secondary = True

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Creating an AlloyDB cluster.")
        mock_get_cluster.assert_called_once()
        assert not mock_create_cluster.called
        mock_create_secondary_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            cluster=TEST_CLUSTER,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_to_dict.assert_called_once_with(mock_operation_result)
        mock_get_operation_result.assert_called_once_with(mock_operation)

        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Cluster.to_dict"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("_get_cluster"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_cluster,
        mock_to_dict,
    ):
        mock_get_cluster.return_value = None
        mock_create_cluster = mock_hook.return_value.create_cluster
        mock_create_secondary_cluster = mock_hook.return_value.create_secondary_cluster
        mock_operation = mock_create_cluster.return_value
        mock_get_operation_result.return_value = None

        mock_context = mock.MagicMock()
        self.operator.validate_request = True

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Validating a Create AlloyDB cluster request.")
        mock_get_cluster.assert_called_once()
        mock_create_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            cluster=TEST_CLUSTER,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_create_secondary_cluster.called
        assert not mock_to_dict.called
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert result is None

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Cluster.to_dict"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("_get_cluster"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request_is_secondary(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_cluster,
        mock_to_dict,
    ):
        mock_get_cluster.return_value = None
        mock_create_cluster = mock_hook.return_value.create_cluster
        mock_create_secondary_cluster = mock_hook.return_value.create_secondary_cluster
        mock_operation = mock_create_secondary_cluster.return_value
        mock_get_operation_result.return_value = None

        mock_context = mock.MagicMock()
        self.operator.validate_request = True
        self.operator.is_secondary = True

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Validating a Create AlloyDB cluster request.")
        mock_get_cluster.assert_called_once()
        mock_create_secondary_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            cluster=TEST_CLUSTER,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_create_cluster.called
        assert not mock_to_dict.called
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert result is None

    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("_get_cluster"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_already_exists(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_cluster,
    ):
        expected_result = mock_get_cluster.return_value
        mock_create_cluster = mock_hook.return_value.create_cluster
        mock_create_secondary_cluster = mock_hook.return_value.create_secondary_cluster

        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        assert not mock_log.info.called
        mock_get_cluster.assert_called_once()
        assert not mock_create_cluster.called
        assert not mock_create_secondary_cluster.called
        assert not mock_get_operation_result.called
        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Cluster.to_dict"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("_get_cluster"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_CLUSTER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_exception(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_cluster,
        mock_to_dict,
    ):
        mock_get_cluster.return_value = None
        mock_create_cluster = mock_hook.return_value.create_cluster
        mock_create_secondary_cluster = mock_hook.return_value.create_secondary_cluster
        mock_create_cluster.side_effect = Exception()
        mock_context = mock.MagicMock()

        with pytest.raises(AirflowException):
            self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Creating an AlloyDB cluster.")
        mock_get_cluster.assert_called_once()
        mock_create_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            cluster=TEST_CLUSTER,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_create_secondary_cluster.called
        assert not mock_to_dict.called
        assert not mock_get_operation_result.called


class TestAlloyDBUpdateClusterOperator:
    def setup_method(self):
        self.operator = AlloyDBUpdateClusterOperator(
            task_id=TEST_TASK_ID,
            cluster_id=TEST_CLUSTER_ID,
            cluster_configuration=TEST_CLUSTER,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.cluster_id == TEST_CLUSTER_ID
        assert self.operator.cluster_configuration == TEST_CLUSTER
        assert self.operator.update_mask == TEST_UPDATE_MASK
        assert self.operator.allow_missing == TEST_ALLOW_MISSING

    def test_template_fields(self):
        expected_template_fields = {"cluster_id", "cluster_configuration", "allow_missing"} | set(
            AlloyDBWriteBaseOperator.template_fields
        )
        assert set(AlloyDBUpdateClusterOperator.template_fields) == expected_template_fields

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Cluster.to_dict"))
    @mock.patch(OPERATOR_MODULE_PATH.format("AlloyDBUpdateClusterOperator.get_operation_result"))
    @mock.patch(OPERATOR_MODULE_PATH.format("AlloyDBUpdateClusterOperator.log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute(self, mock_hook, mock_log, mock_get_operation_result, mock_to_dict):
        mock_update_cluster = mock_hook.return_value.update_cluster
        mock_operation = mock_update_cluster.return_value
        mock_operation_result = mock_get_operation_result.return_value

        expected_result = mock_to_dict.return_value
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        mock_update_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            cluster=TEST_CLUSTER,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_operation_result.assert_called_once_with(mock_operation)
        mock_to_dict.assert_called_once_with(mock_operation_result)
        assert result == expected_result
        mock_log.info.assert_has_calls(
            [
                call("Updating an AlloyDB cluster."),
                call("AlloyDB cluster %s was successfully updated.", TEST_CLUSTER_ID),
            ]
        )

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Cluster.to_dict"))
    @mock.patch(OPERATOR_MODULE_PATH.format("AlloyDBUpdateClusterOperator.get_operation_result"))
    @mock.patch(OPERATOR_MODULE_PATH.format("AlloyDBUpdateClusterOperator.log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_to_dict,
    ):
        mock_update_cluster = mock_hook.return_value.update_cluster
        mock_operation = mock_update_cluster.return_value
        mock_get_operation_result.return_value = None

        expected_message = "Validating an Update AlloyDB cluster request."
        mock_context = mock.MagicMock()
        self.operator.validate_request = True

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with(expected_message)
        mock_update_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            cluster=TEST_CLUSTER,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert not mock_to_dict.called
        assert result is None

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Cluster.to_dict"))
    @mock.patch(OPERATOR_MODULE_PATH.format("AlloyDBUpdateClusterOperator.get_operation_result"))
    @mock.patch(OPERATOR_MODULE_PATH.format("AlloyDBUpdateClusterOperator.log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_exception(self, mock_hook, mock_log, mock_get_operation_result, mock_to_dict):
        mock_update_cluster = mock_hook.return_value.update_cluster
        mock_update_cluster.side_effect = Exception

        mock_context = mock.MagicMock()

        with pytest.raises(AirflowException):
            self.operator.execute(context=mock_context)

        mock_update_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            cluster=TEST_CLUSTER,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_get_operation_result.called
        assert not mock_to_dict.called
        mock_log.info.assert_called_once_with("Updating an AlloyDB cluster.")


class TestAlloyDBDeleteClusterOperator:
    def setup_method(self):
        self.operator = AlloyDBDeleteClusterOperator(
            task_id=TEST_TASK_ID,
            cluster_id=TEST_CLUSTER_ID,
            etag=TEST_ETAG,
            force=TEST_FORCE,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.cluster_id == TEST_CLUSTER_ID
        assert self.operator.etag == TEST_ETAG
        assert self.operator.force == TEST_FORCE

    def test_template_fields(self):
        expected_template_fields = {"cluster_id", "etag", "force"} | set(
            AlloyDBWriteBaseOperator.template_fields
        )
        assert set(AlloyDBDeleteClusterOperator.template_fields) == expected_template_fields

    @mock.patch(OPERATOR_MODULE_PATH.format("AlloyDBDeleteClusterOperator.get_operation_result"))
    @mock.patch(OPERATOR_MODULE_PATH.format("AlloyDBDeleteClusterOperator.log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute(self, mock_hook, mock_log, mock_get_operation_result):
        mock_delete_cluster = mock_hook.return_value.delete_cluster
        mock_operation = mock_delete_cluster.return_value
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        mock_delete_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            etag=TEST_ETAG,
            force=TEST_FORCE,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert result is None
        mock_log.info.assert_has_calls(
            [
                call("Deleting an AlloyDB cluster."),
                call("AlloyDB cluster %s was successfully removed.", TEST_CLUSTER_ID),
            ]
        )

    @mock.patch(OPERATOR_MODULE_PATH.format("AlloyDBDeleteClusterOperator.get_operation_result"))
    @mock.patch(OPERATOR_MODULE_PATH.format("AlloyDBDeleteClusterOperator.log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request(self, mock_hook, mock_log, mock_get_operation_result):
        mock_delete_cluster = mock_hook.return_value.delete_cluster
        mock_operation = mock_delete_cluster.return_value
        mock_context = mock.MagicMock()
        self.operator.validate_request = True

        result = self.operator.execute(context=mock_context)

        mock_delete_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            etag=TEST_ETAG,
            force=TEST_FORCE,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert result is None
        mock_log.info.assert_called_once_with("Validating a Delete AlloyDB cluster request.")

    @mock.patch(OPERATOR_MODULE_PATH.format("AlloyDBDeleteClusterOperator.get_operation_result"))
    @mock.patch(OPERATOR_MODULE_PATH.format("AlloyDBDeleteClusterOperator.log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_exception(self, mock_hook, mock_log, mock_get_operation_result):
        mock_delete_cluster = mock_hook.return_value.delete_cluster
        mock_delete_cluster.side_effect = Exception
        mock_context = mock.MagicMock()

        with pytest.raises(AirflowException):
            _ = self.operator.execute(context=mock_context)

        mock_delete_cluster.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            etag=TEST_ETAG,
            force=TEST_FORCE,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_get_operation_result.called
        mock_log.info.assert_called_once_with("Deleting an AlloyDB cluster.")


class TestAlloyDBCreateInstanceOperator:
    def setup_method(self):
        self.operator = AlloyDBCreateInstanceOperator(
            task_id=TEST_TASK_ID,
            instance_id=TEST_INSTANCE_ID,
            cluster_id=TEST_CLUSTER_ID,
            instance_configuration=TEST_INSTANCE,
            is_secondary=TEST_IS_SECONDARY,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.instance_id == TEST_INSTANCE_ID
        assert self.operator.cluster_id == TEST_CLUSTER_ID
        assert self.operator.instance_configuration == TEST_INSTANCE
        assert self.operator.is_secondary == TEST_IS_SECONDARY

    def test_template_fields(self):
        expected_template_fields = {
            "cluster_id",
            "instance_id",
            "is_secondary",
            "instance_configuration",
        } | set(AlloyDBWriteBaseOperator.template_fields)
        assert set(AlloyDBCreateInstanceOperator.template_fields) == expected_template_fields

    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_get_instance_not_found(self, mock_hook, mock_log):
        mock_get_instance = mock_hook.return_value.get_instance
        mock_get_instance.side_effect = NotFound("Not found")

        result = self.operator._get_instance()

        mock_get_instance.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            instance_id=TEST_INSTANCE_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
        )
        mock_log.info.assert_has_calls(
            [
                call("Checking if the instance %s exists already...", TEST_INSTANCE_ID),
                call("The instance %s does not exist yet.", TEST_INSTANCE_ID),
            ]
        )
        assert result is None

    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_get_instance_exception(self, mock_hook, mock_log):
        mock_get_instance = mock_hook.return_value.get_instance
        mock_get_instance.side_effect = Exception("Test exception")

        with pytest.raises(AirflowException):
            self.operator._get_instance()

        mock_get_instance.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            instance_id=TEST_INSTANCE_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
        )
        mock_log.info.assert_called_once_with(
            "Checking if the instance %s exists already...", TEST_INSTANCE_ID
        )

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Instance.to_dict"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_get_instance(self, mock_hook, mock_log, mock_to_dict):
        mock_get_instance = mock_hook.return_value.get_instance
        mock_instance = mock_get_instance.return_value
        expected_result = mock_to_dict.return_value

        result = self.operator._get_instance()

        mock_get_instance.assert_called_once_with(
            instance_id=TEST_INSTANCE_ID,
            cluster_id=TEST_CLUSTER_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
        )
        mock_log.info.assert_has_calls(
            [
                call("Checking if the instance %s exists already...", TEST_INSTANCE_ID),
                call(
                    "AlloyDB instance %s already exists in the cluster %s.", TEST_CLUSTER_ID, TEST_INSTANCE_ID
                ),
            ]
        )
        mock_to_dict.assert_called_once_with(mock_instance)
        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Instance.to_dict"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("_get_instance"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_instance,
        mock_to_dict,
    ):
        mock_get_instance.return_value = None
        mock_create_instance = mock_hook.return_value.create_instance
        mock_create_secondary_instance = mock_hook.return_value.create_secondary_instance
        mock_operation = mock_create_instance.return_value
        mock_operation_result = mock_get_operation_result.return_value

        expected_result = mock_to_dict.return_value
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Creating an AlloyDB instance.")
        mock_get_instance.assert_called_once()
        mock_create_instance.assert_called_once_with(
            instance_id=TEST_INSTANCE_ID,
            cluster_id=TEST_CLUSTER_ID,
            instance=TEST_INSTANCE,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_create_secondary_instance.called
        mock_to_dict.assert_called_once_with(mock_operation_result)
        mock_get_operation_result.assert_called_once_with(mock_operation)

        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Instance.to_dict"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("_get_instance"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_is_secondary(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_instance,
        mock_to_dict,
    ):
        mock_get_instance.return_value = None
        mock_create_instance = mock_hook.return_value.create_instance
        mock_create_secondary_instance = mock_hook.return_value.create_secondary_instance
        mock_operation = mock_create_secondary_instance.return_value
        mock_operation_result = mock_get_operation_result.return_value

        expected_result = mock_to_dict.return_value
        mock_context = mock.MagicMock()
        self.operator.is_secondary = True

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Creating an AlloyDB instance.")
        mock_get_instance.assert_called_once()
        assert not mock_create_instance.called
        mock_create_secondary_instance.assert_called_once_with(
            instance_id=TEST_INSTANCE_ID,
            cluster_id=TEST_CLUSTER_ID,
            instance=TEST_INSTANCE,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_to_dict.assert_called_once_with(mock_operation_result)
        mock_get_operation_result.assert_called_once_with(mock_operation)

        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Instance.to_dict"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("_get_instance"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_instance,
        mock_to_dict,
    ):
        mock_get_instance.return_value = None
        mock_create_instance = mock_hook.return_value.create_instance
        mock_create_secondary_instance = mock_hook.return_value.create_secondary_instance
        mock_operation = mock_create_instance.return_value
        mock_get_operation_result.return_value = None

        mock_context = mock.MagicMock()
        self.operator.validate_request = True

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Validating a Create AlloyDB instance request.")
        mock_get_instance.assert_called_once()
        mock_create_instance.assert_called_once_with(
            instance_id=TEST_INSTANCE_ID,
            cluster_id=TEST_CLUSTER_ID,
            instance=TEST_INSTANCE,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_create_secondary_instance.called
        assert not mock_to_dict.called
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert result is None

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Instance.to_dict"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("_get_instance"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request_is_secondary(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_instance,
        mock_to_dict,
    ):
        mock_get_instance.return_value = None
        mock_create_instance = mock_hook.return_value.create_instance
        mock_create_secondary_instance = mock_hook.return_value.create_secondary_instance
        mock_operation = mock_create_secondary_instance.return_value
        mock_get_operation_result.return_value = None

        mock_context = mock.MagicMock()
        self.operator.validate_request = True
        self.operator.is_secondary = True

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Validating a Create AlloyDB instance request.")
        mock_get_instance.assert_called_once()
        mock_create_secondary_instance.assert_called_once_with(
            instance_id=TEST_INSTANCE_ID,
            cluster_id=TEST_CLUSTER_ID,
            instance=TEST_INSTANCE,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_create_instance.called
        assert not mock_to_dict.called
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert result is None

    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("_get_instance"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_already_exists(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_instance,
    ):
        expected_result = mock_get_instance.return_value
        mock_create_instance = mock_hook.return_value.create_instance
        mock_create_secondary_instance = mock_hook.return_value.create_secondary_instance

        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        assert not mock_log.info.called
        mock_get_instance.assert_called_once()
        assert not mock_create_instance.called
        assert not mock_create_secondary_instance.called
        assert not mock_get_operation_result.called
        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Instance.to_dict"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("_get_instance"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_exception(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_instance,
        mock_to_dict,
    ):
        mock_get_instance.return_value = None
        mock_create_instance = mock_hook.return_value.create_instance
        mock_create_secondary_instance = mock_hook.return_value.create_secondary_instance
        mock_create_instance.side_effect = Exception()
        mock_context = mock.MagicMock()

        with pytest.raises(AirflowException):
            self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Creating an AlloyDB instance.")
        mock_get_instance.assert_called_once()
        mock_create_instance.assert_called_once_with(
            instance_id=TEST_INSTANCE_ID,
            cluster_id=TEST_CLUSTER_ID,
            instance=TEST_INSTANCE,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_create_secondary_instance.called
        assert not mock_to_dict.called
        assert not mock_get_operation_result.called


class TestAlloyDBUpdateInstanceOperator:
    def setup_method(self):
        self.operator = AlloyDBUpdateInstanceOperator(
            task_id=TEST_TASK_ID,
            instance_id=TEST_INSTANCE_ID,
            cluster_id=TEST_CLUSTER_ID,
            instance_configuration=TEST_INSTANCE,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.instance_id == TEST_INSTANCE_ID
        assert self.operator.cluster_id == TEST_CLUSTER_ID
        assert self.operator.instance_configuration == TEST_INSTANCE
        assert self.operator.update_mask == TEST_UPDATE_MASK
        assert self.operator.allow_missing == TEST_ALLOW_MISSING

    def test_template_fields(self):
        expected_template_fields = {
            "cluster_id",
            "instance_id",
            "instance_configuration",
            "update_mask",
            "allow_missing",
        } | set(AlloyDBWriteBaseOperator.template_fields)
        assert set(AlloyDBUpdateInstanceOperator.template_fields) == expected_template_fields

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Instance.to_dict"))
    @mock.patch(UPDATE_INSTANCE_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(UPDATE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute(self, mock_hook, mock_log, mock_get_operation_result, mock_to_dict):
        mock_update_instance = mock_hook.return_value.update_instance
        mock_operation = mock_update_instance.return_value
        mock_operation_result = mock_get_operation_result.return_value

        expected_result = mock_to_dict.return_value
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        mock_update_instance.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            instance_id=TEST_INSTANCE_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            instance=TEST_INSTANCE,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_operation_result.assert_called_once_with(mock_operation)
        mock_to_dict.assert_called_once_with(mock_operation_result)
        assert result == expected_result
        mock_log.info.assert_has_calls(
            [
                call("Updating an AlloyDB instance."),
                call("AlloyDB instance %s was successfully updated.", TEST_CLUSTER_ID),
            ]
        )

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Cluster.to_dict"))
    @mock.patch(UPDATE_INSTANCE_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(UPDATE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_to_dict,
    ):
        mock_update_instance = mock_hook.return_value.update_instance
        mock_operation = mock_update_instance.return_value
        mock_get_operation_result.return_value = None

        expected_message = "Validating an Update AlloyDB instance request."
        mock_context = mock.MagicMock()
        self.operator.validate_request = True

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with(expected_message)
        mock_update_instance.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            instance_id=TEST_INSTANCE_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            instance=TEST_INSTANCE,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert not mock_to_dict.called
        assert result is None

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Cluster.to_dict"))
    @mock.patch(UPDATE_INSTANCE_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(UPDATE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_exception(self, mock_hook, mock_log, mock_get_operation_result, mock_to_dict):
        mock_update_instance = mock_hook.return_value.update_instance
        mock_update_instance.side_effect = Exception

        mock_context = mock.MagicMock()

        with pytest.raises(AirflowException):
            self.operator.execute(context=mock_context)

        mock_update_instance.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            instance_id=TEST_INSTANCE_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            instance=TEST_INSTANCE,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_get_operation_result.called
        assert not mock_to_dict.called
        mock_log.info.assert_called_once_with("Updating an AlloyDB instance.")


class TestAlloyDBDeleteInstanceOperator:
    def setup_method(self):
        self.operator = AlloyDBDeleteInstanceOperator(
            task_id=TEST_TASK_ID,
            instance_id=TEST_INSTANCE_ID,
            cluster_id=TEST_CLUSTER_ID,
            etag=TEST_ETAG,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.cluster_id == TEST_CLUSTER_ID
        assert self.operator.instance_id == TEST_INSTANCE_ID
        assert self.operator.etag == TEST_ETAG

    def test_template_fields(self):
        expected_template_fields = {"cluster_id", "instance_id", "etag"} | set(
            AlloyDBWriteBaseOperator.template_fields
        )
        assert set(AlloyDBDeleteInstanceOperator.template_fields) == expected_template_fields

    @mock.patch(DELETE_INSTANCE_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(DELETE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute(self, mock_hook, mock_log, mock_get_operation_result):
        mock_delete_instance = mock_hook.return_value.delete_instance
        mock_operation = mock_delete_instance.return_value
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        mock_delete_instance.assert_called_once_with(
            instance_id=TEST_INSTANCE_ID,
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            etag=TEST_ETAG,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert result is None
        mock_log.info.assert_has_calls(
            [
                call("Deleting an AlloyDB instance."),
                call("AlloyDB instance %s was successfully removed.", TEST_INSTANCE_ID),
            ]
        )

    @mock.patch(DELETE_INSTANCE_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(DELETE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request(self, mock_hook, mock_log, mock_get_operation_result):
        mock_delete_instance = mock_hook.return_value.delete_instance
        mock_operation = mock_delete_instance.return_value
        mock_context = mock.MagicMock()
        self.operator.validate_request = True

        result = self.operator.execute(context=mock_context)

        mock_delete_instance.assert_called_once_with(
            instance_id=TEST_INSTANCE_ID,
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            etag=TEST_ETAG,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert result is None
        mock_log.info.assert_called_once_with("Validating a Delete AlloyDB instance request.")

    @mock.patch(DELETE_INSTANCE_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(DELETE_INSTANCE_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_exception(self, mock_hook, mock_log, mock_get_operation_result):
        mock_delete_instance = mock_hook.return_value.delete_instance
        mock_delete_instance.side_effect = Exception
        mock_context = mock.MagicMock()

        with pytest.raises(AirflowException):
            _ = self.operator.execute(context=mock_context)

        mock_delete_instance.assert_called_once_with(
            instance_id=TEST_INSTANCE_ID,
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            etag=TEST_ETAG,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_get_operation_result.called
        mock_log.info.assert_called_once_with("Deleting an AlloyDB instance.")


class TestAlloyDBCreateUserOperator:
    def setup_method(self):
        self.operator = AlloyDBCreateUserOperator(
            task_id=TEST_TASK_ID,
            user_id=TEST_USER_ID,
            user_configuration=TEST_USER,
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.user_id == TEST_USER_ID
        assert self.operator.user_configuration == TEST_USER
        assert self.operator.cluster_id == TEST_CLUSTER_ID

    def test_template_fields(self):
        expected_template_fields = {
            "cluster_id",
            "user_id",
            "user_configuration",
        } | set(AlloyDBWriteBaseOperator.template_fields)
        assert set(AlloyDBCreateUserOperator.template_fields) == expected_template_fields

    @mock.patch(CREATE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_get_user_not_found(self, mock_hook, mock_log):
        mock_get_user = mock_hook.return_value.get_user
        mock_get_user.side_effect = NotFound("Not found")

        result = self.operator._get_user()

        mock_get_user.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            user_id=TEST_USER_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
        )
        mock_log.info.assert_has_calls(
            [
                call("Checking if the user %s exists already...", TEST_USER_ID),
                call("The user %s does not exist yet.", TEST_USER_ID),
            ]
        )
        assert result is None

    @mock.patch(CREATE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_get_user_exception(self, mock_hook, mock_log):
        mock_get_user = mock_hook.return_value.get_user
        mock_get_user.side_effect = Exception("Test exception")

        with pytest.raises(AirflowException):
            self.operator._get_user()

        mock_get_user.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            user_id=TEST_USER_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
        )
        mock_log.info.assert_called_once_with("Checking if the user %s exists already...", TEST_USER_ID)

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.User.to_dict"))
    @mock.patch(CREATE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_get_user(self, mock_hook, mock_log, mock_to_dict):
        mock_get_user = mock_hook.return_value.get_user
        mock_user = mock_get_user.return_value
        expected_result = mock_to_dict.return_value

        result = self.operator._get_user()

        mock_get_user.assert_called_once_with(
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
        )
        mock_log.info.assert_has_calls(
            [
                call("Checking if the user %s exists already...", TEST_USER_ID),
                call("AlloyDB user %s already exists in the cluster %s.", TEST_USER_ID, TEST_CLUSTER_ID),
            ]
        )
        mock_to_dict.assert_called_once_with(mock_user)
        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.User.to_dict"))
    @mock.patch(CREATE_USER_OPERATOR_PATH.format("_get_user"))
    @mock.patch(CREATE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute(self, mock_hook, mock_log, mock_get_user, mock_to_dict):
        mock_get_user.return_value = None
        mock_create_user = mock_hook.return_value.create_user
        mock_user = mock_create_user.return_value

        expected_result = mock_to_dict.return_value
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_has_calls(
            [
                call("Creating an AlloyDB user."),
                call("AlloyDB user %s was successfully created.", TEST_USER_ID),
            ]
        )
        mock_get_user.assert_called_once()
        mock_create_user.assert_called_once_with(
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            user=TEST_USER,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_to_dict.assert_called_once_with(mock_user)
        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.User.to_dict"))
    @mock.patch(CREATE_USER_OPERATOR_PATH.format("_get_user"))
    @mock.patch(CREATE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request(self, mock_hook, mock_log, mock_get_user, mock_to_dict):
        mock_get_user.return_value = None
        mock_create_user = mock_hook.return_value.create_user

        mock_context = mock.MagicMock()
        self.operator.validate_request = True

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Validating a Create AlloyDB user request.")
        mock_get_user.assert_called_once()
        mock_create_user.assert_called_once_with(
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            user=TEST_USER,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_to_dict.called
        assert result is None

    @mock.patch(CREATE_USER_OPERATOR_PATH.format("_get_user"))
    @mock.patch(CREATE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_already_exists(self, mock_hook, mock_log, mock_get_user):
        expected_result = mock_get_user.return_value
        mock_create_user = mock_hook.return_value.create_user
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        assert not mock_log.info.called
        mock_get_user.assert_called_once()
        assert not mock_create_user.called
        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.User.to_dict"))
    @mock.patch(CREATE_USER_OPERATOR_PATH.format("_get_user"))
    @mock.patch(CREATE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_exception(self, mock_hook, mock_log, mock_get_user, mock_to_dict):
        mock_get_user.return_value = None
        mock_create_user = mock_hook.return_value.create_user
        mock_create_user.side_effect = Exception()
        mock_context = mock.MagicMock()

        with pytest.raises(AirflowException):
            self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Creating an AlloyDB user.")
        mock_get_user.assert_called_once()
        mock_create_user.assert_called_once_with(
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            user=TEST_INSTANCE,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_to_dict.called


class TestAlloyDBUpdateUserOperator:
    def setup_method(self):
        self.operator = AlloyDBUpdateUserOperator(
            task_id=TEST_TASK_ID,
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            user_configuration=TEST_USER,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.user_id == TEST_USER_ID
        assert self.operator.cluster_id == TEST_CLUSTER_ID
        assert self.operator.user_configuration == TEST_USER
        assert self.operator.update_mask == TEST_UPDATE_MASK
        assert self.operator.allow_missing == TEST_ALLOW_MISSING

    def test_template_fields(self):
        expected_template_fields = {
            "cluster_id",
            "user_id",
            "user_configuration",
            "update_mask",
            "allow_missing",
        } | set(AlloyDBWriteBaseOperator.template_fields)
        assert set(AlloyDBUpdateUserOperator.template_fields) == expected_template_fields

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.User.to_dict"))
    @mock.patch(UPDATE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute(self, mock_hook, mock_log, mock_to_dict):
        mock_update_user = mock_hook.return_value.update_user
        mock_user = mock_update_user.return_value
        expected_result = mock_to_dict.return_value
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        mock_update_user.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            user_id=TEST_USER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            user=TEST_USER,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_to_dict.assert_called_once_with(mock_user)
        assert result == expected_result
        mock_log.info.assert_has_calls(
            [
                call("Updating an AlloyDB user."),
                call("AlloyDB user %s was successfully updated.", TEST_USER_ID),
            ]
        )

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.User.to_dict"))
    @mock.patch(UPDATE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request(self, mock_hook, mock_log, mock_to_dict):
        mock_update_user = mock_hook.return_value.update_user

        expected_message = "Validating an Update AlloyDB user request."
        mock_context = mock.MagicMock()
        self.operator.validate_request = True

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with(expected_message)
        mock_update_user.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            user_id=TEST_USER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            user=TEST_USER,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_to_dict.called
        assert result is None

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.User.to_dict"))
    @mock.patch(UPDATE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_exception(self, mock_hook, mock_log, mock_to_dict):
        mock_update_user = mock_hook.return_value.update_user
        mock_update_user.side_effect = Exception

        mock_context = mock.MagicMock()

        with pytest.raises(AirflowException):
            self.operator.execute(context=mock_context)

        mock_update_user.assert_called_once_with(
            cluster_id=TEST_CLUSTER_ID,
            user_id=TEST_USER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            user=TEST_USER,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_to_dict.called
        mock_log.info.assert_called_once_with("Updating an AlloyDB user.")


class TestAlloyDBDeleteUserOperator:
    def setup_method(self):
        self.operator = AlloyDBDeleteUserOperator(
            task_id=TEST_TASK_ID,
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.cluster_id == TEST_CLUSTER_ID
        assert self.operator.user_id == TEST_USER_ID

    def test_template_fields(self):
        expected_template_fields = {"user_id", "cluster_id"} | set(AlloyDBWriteBaseOperator.template_fields)
        assert set(AlloyDBDeleteUserOperator.template_fields) == expected_template_fields

    @mock.patch(DELETE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute(self, mock_hook, mock_log):
        mock_delete_user = mock_hook.return_value.delete_user
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        mock_delete_user.assert_called_once_with(
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert result is None
        mock_log.info.assert_has_calls(
            [
                call("Deleting an AlloyDB user."),
                call("AlloyDB user %s was successfully removed.", TEST_USER_ID),
            ]
        )

    @mock.patch(DELETE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request(self, mock_hook, mock_log):
        mock_delete_user = mock_hook.return_value.delete_user
        mock_context = mock.MagicMock()
        self.operator.validate_request = True

        result = self.operator.execute(context=mock_context)

        mock_delete_user.assert_called_once_with(
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert result is None
        mock_log.info.assert_called_once_with("Validating a Delete AlloyDB user request.")

    @mock.patch(DELETE_USER_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_exception(self, mock_hook, mock_log):
        mock_delete_user = mock_hook.return_value.delete_user
        mock_delete_user.side_effect = Exception
        mock_context = mock.MagicMock()

        with pytest.raises(AirflowException):
            self.operator.execute(context=mock_context)

        mock_delete_user.assert_called_once_with(
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_log.info.assert_called_once_with("Deleting an AlloyDB user.")


class TestAlloyDBCreateBackupOperator:
    def setup_method(self):
        self.operator = AlloyDBCreateBackupOperator(
            task_id=TEST_TASK_ID,
            backup_id=TEST_BACKUP_ID,
            backup_configuration=TEST_BACKUP,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.backup_id == TEST_BACKUP_ID
        assert self.operator.backup_configuration == TEST_BACKUP

    def test_template_fields(self):
        expected_template_fields = {
            "backup_id",
            "backup_configuration",
        } | set(AlloyDBWriteBaseOperator.template_fields)
        assert set(AlloyDBCreateBackupOperator.template_fields) == expected_template_fields

    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_get_backup_not_found(self, mock_hook, mock_log):
        mock_get_backup = mock_hook.return_value.get_backup
        mock_get_backup.side_effect = NotFound("Not found")

        result = self.operator._get_backup()

        mock_get_backup.assert_called_once_with(
            backup_id=TEST_BACKUP_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
        )
        mock_log.info.assert_has_calls(
            [
                call("Checking if the backup %s exists already...", TEST_BACKUP_ID),
                call("The backup %s does not exist yet.", TEST_BACKUP_ID),
            ]
        )
        assert result is None

    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_get_backup_exception(self, mock_hook, mock_log):
        mock_get_backup = mock_hook.return_value.get_backup
        mock_get_backup.side_effect = Exception("Test exception")

        with pytest.raises(AirflowException):
            self.operator._get_backup()

        mock_get_backup.assert_called_once_with(
            backup_id=TEST_BACKUP_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
        )
        mock_log.info.assert_called_once_with("Checking if the backup %s exists already...", TEST_BACKUP_ID)

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Backup.to_dict"))
    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_get_backup(self, mock_hook, mock_log, mock_to_dict):
        mock_get_backup = mock_hook.return_value.get_backup
        mock_instance = mock_get_backup.return_value
        expected_result = mock_to_dict.return_value

        result = self.operator._get_backup()

        mock_get_backup.assert_called_once_with(
            backup_id=TEST_BACKUP_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
        )
        mock_log.info.assert_has_calls(
            [
                call("Checking if the backup %s exists already...", TEST_BACKUP_ID),
                call("AlloyDB backup %s already exists.", TEST_BACKUP_ID),
            ]
        )
        mock_to_dict.assert_called_once_with(mock_instance)
        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Backup.to_dict"))
    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("_get_backup"))
    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_backup,
        mock_to_dict,
    ):
        mock_get_backup.return_value = None
        mock_create_backup = mock_hook.return_value.create_backup
        mock_operation = mock_create_backup.return_value
        mock_operation_result = mock_get_operation_result.return_value

        expected_result = mock_to_dict.return_value
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Creating an AlloyDB backup.")
        mock_get_backup.assert_called_once()
        mock_create_backup.assert_called_once_with(
            backup_id=TEST_BACKUP_ID,
            backup=TEST_BACKUP,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_to_dict.assert_called_once_with(mock_operation_result)
        mock_get_operation_result.assert_called_once_with(mock_operation)

        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Backup.to_dict"))
    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("_get_backup"))
    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_backup,
        mock_to_dict,
    ):
        mock_get_backup.return_value = None
        mock_create_backup = mock_hook.return_value.create_backup
        mock_operation = mock_create_backup.return_value
        mock_get_operation_result.return_value = None

        mock_context = mock.MagicMock()
        self.operator.validate_request = True

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Validating a Create AlloyDB backup request.")
        mock_get_backup.assert_called_once()
        mock_create_backup.assert_called_once_with(
            backup_id=TEST_BACKUP_ID,
            backup=TEST_BACKUP,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_to_dict.called
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert result is None

    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("_get_backup"))
    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_already_exists(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_backup,
    ):
        expected_result = mock_get_backup.return_value
        mock_create_instance = mock_hook.return_value.create_instance

        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        assert not mock_log.info.called
        mock_get_backup.assert_called_once()
        assert not mock_create_instance.called
        assert not mock_get_operation_result.called
        assert result == expected_result

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Backup.to_dict"))
    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("_get_backup"))
    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(CREATE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_exception(
        self,
        mock_hook,
        mock_log,
        mock_get_operation_result,
        mock_get_backup,
        mock_to_dict,
    ):
        mock_get_backup.return_value = None
        mock_create_backup = mock_hook.return_value.create_backup
        mock_create_backup.side_effect = Exception()
        mock_context = mock.MagicMock()

        with pytest.raises(AirflowException):
            self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with("Creating an AlloyDB backup.")
        mock_get_backup.assert_called_once()
        mock_create_backup.assert_called_once_with(
            backup_id=TEST_BACKUP_ID,
            backup=TEST_BACKUP,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_to_dict.called
        assert not mock_get_operation_result.called


class TestAlloyDBUpdateBackupOperator:
    def setup_method(self):
        self.operator = AlloyDBUpdateBackupOperator(
            task_id=TEST_TASK_ID,
            backup_id=TEST_BACKUP_ID,
            backup_configuration=TEST_BACKUP,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.backup_id == TEST_BACKUP_ID
        assert self.operator.backup_configuration == TEST_BACKUP
        assert self.operator.update_mask == TEST_UPDATE_MASK
        assert self.operator.allow_missing == TEST_ALLOW_MISSING

    def test_template_fields(self):
        expected_template_fields = {
            "backup_id",
            "backup_configuration",
            "update_mask",
            "allow_missing",
        } | set(AlloyDBWriteBaseOperator.template_fields)
        assert set(AlloyDBUpdateBackupOperator.template_fields) == expected_template_fields

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Backup.to_dict"))
    @mock.patch(UPDATE_BACKUP_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(UPDATE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute(self, mock_hook, mock_log, mock_get_operation_result, mock_to_dict):
        mock_update_backup = mock_hook.return_value.update_backup
        mock_operation = mock_update_backup.return_value
        mock_operation_result = mock_get_operation_result.return_value

        expected_result = mock_to_dict.return_value
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        mock_update_backup.assert_called_once_with(
            backup_id=TEST_BACKUP_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            backup=TEST_BACKUP,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_operation_result.assert_called_once_with(mock_operation)
        mock_to_dict.assert_called_once_with(mock_operation_result)
        assert result == expected_result
        mock_log.info.assert_has_calls(
            [
                call("Updating an AlloyDB backup."),
                call("AlloyDB backup %s was successfully updated.", TEST_BACKUP_ID),
            ]
        )

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Backup.to_dict"))
    @mock.patch(UPDATE_BACKUP_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(UPDATE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request(self, mock_hook, mock_log, mock_get_operation_result, mock_to_dict):
        mock_update_ackup = mock_hook.return_value.update_backup
        mock_operation = mock_update_ackup.return_value
        mock_get_operation_result.return_value = None

        expected_message = "Validating an Update AlloyDB backup request."
        mock_context = mock.MagicMock()
        self.operator.validate_request = True

        result = self.operator.execute(context=mock_context)

        mock_log.info.assert_called_once_with(expected_message)
        mock_update_ackup.assert_called_once_with(
            backup_id=TEST_BACKUP_ID,
            backup=TEST_BACKUP,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert not mock_to_dict.called
        assert result is None

    @mock.patch(OPERATOR_MODULE_PATH.format("alloydb_v1.Backup.to_dict"))
    @mock.patch(UPDATE_BACKUP_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(UPDATE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_exception(self, mock_hook, mock_log, mock_get_operation_result, mock_to_dict):
        mock_update_backup = mock_hook.return_value.update_backup
        mock_update_backup.side_effect = Exception

        mock_context = mock.MagicMock()

        with pytest.raises(AirflowException):
            self.operator.execute(context=mock_context)

        mock_update_backup.assert_called_once_with(
            backup_id=TEST_BACKUP_ID,
            backup=TEST_BACKUP,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            update_mask=TEST_UPDATE_MASK,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_get_operation_result.called
        assert not mock_to_dict.called
        mock_log.info.assert_called_once_with("Updating an AlloyDB backup.")


class TestAlloyDBDeleteBackupOperator:
    def setup_method(self):
        self.operator = AlloyDBDeleteBackupOperator(
            task_id=TEST_TASK_ID,
            backup_id=TEST_BACKUP_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            gcp_conn_id=TEST_GCP_CONN_ID,
            request_id=TEST_REQUEST_ID,
            validate_request=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

    def test_init(self):
        assert self.operator.backup_id == TEST_BACKUP_ID

    def test_template_fields(self):
        expected_template_fields = {"backup_id"} | set(AlloyDBWriteBaseOperator.template_fields)
        assert set(AlloyDBDeleteBackupOperator.template_fields) == expected_template_fields

    @mock.patch(DELETE_BACKUP_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(DELETE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute(self, mock_hook, mock_log, mock_get_operation_result):
        mock_delete_backup = mock_hook.return_value.delete_backup
        mock_operation = mock_delete_backup.return_value
        mock_context = mock.MagicMock()

        result = self.operator.execute(context=mock_context)

        mock_delete_backup.assert_called_once_with(
            backup_id=TEST_BACKUP_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert result is None
        mock_log.info.assert_has_calls(
            [
                call("Deleting an AlloyDB backup."),
                call("AlloyDB backup %s was successfully removed.", TEST_BACKUP_ID),
            ]
        )

    @mock.patch(DELETE_BACKUP_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(DELETE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_validate_request(self, mock_hook, mock_log, mock_get_operation_result):
        mock_delete_backup = mock_hook.return_value.delete_backup
        mock_operation = mock_delete_backup.return_value
        mock_context = mock.MagicMock()
        self.operator.validate_request = True

        result = self.operator.execute(context=mock_context)

        mock_delete_backup.assert_called_once_with(
            backup_id=TEST_BACKUP_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            request_id=TEST_REQUEST_ID,
            validate_only=True,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_operation_result.assert_called_once_with(mock_operation)
        assert result is None
        mock_log.info.assert_called_once_with("Validating a Delete AlloyDB backup request.")

    @mock.patch(DELETE_BACKUP_OPERATOR_PATH.format("get_operation_result"))
    @mock.patch(DELETE_BACKUP_OPERATOR_PATH.format("log"))
    @mock.patch(ALLOY_DB_HOOK_PATH, new_callable=mock.PropertyMock)
    def test_execute_exception(self, mock_hook, mock_log, mock_get_operation_result):
        mock_delete_backup = mock_hook.return_value.delete_backup
        mock_delete_backup.side_effect = Exception
        mock_context = mock.MagicMock()

        with pytest.raises(AirflowException):
            _ = self.operator.execute(context=mock_context)

        mock_delete_backup.assert_called_once_with(
            backup_id=TEST_BACKUP_ID,
            project_id=TEST_GCP_PROJECT,
            location=TEST_GCP_REGION,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert not mock_get_operation_result.called
        mock_log.info.assert_called_once_with("Deleting an AlloyDB backup.")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.marketing_platform.hooks.campaign_manager import GoogleCampaignManagerHook

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_default_project_id

API_VERSION = "v4"
GCP_CONN_ID = "google_cloud_default"

REPORT_ID = "REPORT_ID"
PROFILE_ID = "PROFILE_ID"
ENCRYPTION_SOURCE = "encryption_source"
ENCRYPTION_ENTITY_TYPE = "encryption_entity_type"
ENCRYPTION_ENTITY_ID = 1234567


class TestGoogleCampaignManagerHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = GoogleCampaignManagerHook(gcp_conn_id=GCP_CONN_ID, api_version=API_VERSION)

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "campaign_manager.GoogleCampaignManagerHook._authorize"
    )
    @mock.patch("airflow.providers.google.marketing_platform.hooks.campaign_manager.build")
    def test_gen_conn(self, mock_build, mock_authorize):
        result = self.hook.get_conn()
        mock_build.assert_called_once_with(
            "dfareporting",
            API_VERSION,
            http=mock_authorize.return_value,
            cache_discovery=False,
        )
        assert mock_build.return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "campaign_manager.GoogleCampaignManagerHook.get_conn"
    )
    def test_delete_report(self, get_conn_mock):
        return_value = "TEST"
        get_conn_mock.return_value.reports.return_value.delete.return_value.execute.return_value = (
            return_value
        )

        result = self.hook.delete_report(profile_id=PROFILE_ID, report_id=REPORT_ID)

        get_conn_mock.return_value.reports.return_value.delete.assert_called_once_with(
            profileId=PROFILE_ID, reportId=REPORT_ID
        )

        assert return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "campaign_manager.GoogleCampaignManagerHook.get_conn"
    )
    def test_get_report(self, get_conn_mock):
        file_id = "FILE_ID"

        return_value = "TEST"
        # fmt: off
        get_conn_mock.return_value.reports.return_value.files.return_value. \
            get.return_value.execute.return_value = return_value
        # fmt: on

        result = self.hook.get_report(profile_id=PROFILE_ID, report_id=REPORT_ID, file_id=file_id)

        get_conn_mock.return_value.reports.return_value.files.return_value.get.assert_called_once_with(
            profileId=PROFILE_ID, reportId=REPORT_ID, fileId=file_id
        )

        assert return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "campaign_manager.GoogleCampaignManagerHook.get_conn"
    )
    def test_get_report_file(self, get_conn_mock):
        file_id = "FILE_ID"

        return_value = "TEST"
        get_conn_mock.return_value.reports.return_value.files.return_value.get_media.return_value = (
            return_value
        )

        result = self.hook.get_report_file(profile_id=PROFILE_ID, report_id=REPORT_ID, file_id=file_id)

        get_conn_mock.return_value.reports.return_value.files.return_value.get_media.assert_called_once_with(
            profileId=PROFILE_ID, reportId=REPORT_ID, fileId=file_id
        )

        assert return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "campaign_manager.GoogleCampaignManagerHook.get_conn"
    )
    def test_insert_report(self, get_conn_mock):
        report = {"body": "test"}

        return_value = "TEST"
        get_conn_mock.return_value.reports.return_value.insert.return_value.execute.return_value = (
            return_value
        )

        result = self.hook.insert_report(profile_id=PROFILE_ID, report=report)

        get_conn_mock.return_value.reports.return_value.insert.assert_called_once_with(
            profileId=PROFILE_ID, body=report
        )

        assert return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "campaign_manager.GoogleCampaignManagerHook.get_conn"
    )
    def test_list_reports(self, get_conn_mock):
        max_results = 42
        scope = "SCOPE"
        sort_field = "SORT_FIELD"
        sort_order = "SORT_ORDER"
        items = ["item"]

        return_value = {"nextPageToken": None, "items": items}
        get_conn_mock.return_value.reports.return_value.list.return_value.execute.return_value = return_value

        request_mock = mock.MagicMock()
        request_mock.execute.return_value = {"nextPageToken": None, "items": items}
        get_conn_mock.return_value.reports.return_value.list_next.side_effect = [
            request_mock,
            request_mock,
            request_mock,
            None,
        ]

        result = self.hook.list_reports(
            profile_id=PROFILE_ID,
            max_results=max_results,
            scope=scope,
            sort_field=sort_field,
            sort_order=sort_order,
        )

        get_conn_mock.return_value.reports.return_value.list.assert_called_once_with(
            profileId=PROFILE_ID,
            maxResults=max_results,
            scope=scope,
            sortField=sort_field,
            sortOrder=sort_order,
        )

        assert items * 4 == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "campaign_manager.GoogleCampaignManagerHook.get_conn"
    )
    def test_patch_report(self, get_conn_mock):
        update_mask = {"test": "test"}

        return_value = "TEST"
        get_conn_mock.return_value.reports.return_value.patch.return_value.execute.return_value = return_value

        result = self.hook.patch_report(profile_id=PROFILE_ID, report_id=REPORT_ID, update_mask=update_mask)

        get_conn_mock.return_value.reports.return_value.patch.assert_called_once_with(
            profileId=PROFILE_ID, reportId=REPORT_ID, body=update_mask
        )

        assert return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "campaign_manager.GoogleCampaignManagerHook.get_conn"
    )
    def test_run_report(self, get_conn_mock):
        synchronous = True

        return_value = "TEST"
        get_conn_mock.return_value.reports.return_value.run.return_value.execute.return_value = return_value

        result = self.hook.run_report(profile_id=PROFILE_ID, report_id=REPORT_ID, synchronous=synchronous)

        get_conn_mock.return_value.reports.return_value.run.assert_called_once_with(
            profileId=PROFILE_ID, reportId=REPORT_ID, synchronous=synchronous
        )

        assert return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "campaign_manager.GoogleCampaignManagerHook.get_conn"
    )
    def test_update_report(self, get_conn_mock):
        return_value = "TEST"
        get_conn_mock.return_value.reports.return_value.update.return_value.execute.return_value = (
            return_value
        )

        result = self.hook.update_report(profile_id=PROFILE_ID, report_id=REPORT_ID)

        get_conn_mock.return_value.reports.return_value.update.assert_called_once_with(
            profileId=PROFILE_ID, reportId=REPORT_ID
        )

        assert return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform."
        "hooks.campaign_manager.GoogleCampaignManagerHook.get_conn"
    )
    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks.campaign_manager.GoogleCampaignManagerHook"
        "._conversions_batch_request"
    )
    def test_conversion_batch_insert(self, batch_request_mock, get_conn_mock):
        conversions = [{"conversions1": "value"}, {"conversions2": "value"}]

        return_value = {"hasFailures": False}
        get_conn_mock.return_value.conversions.return_value.batchinsert.return_value.execute.return_value = (
            return_value
        )

        batch_request_mock.return_value = "batch_request_mock"

        result = self.hook.conversions_batch_insert(
            profile_id=PROFILE_ID,
            conversions=conversions,
            encryption_entity_id=ENCRYPTION_ENTITY_ID,
            encryption_entity_type=ENCRYPTION_ENTITY_TYPE,
            encryption_source=ENCRYPTION_SOURCE,
        )

        batch_request_mock.assert_called_once_with(
            conversions=conversions,
            encryption_entity_id=ENCRYPTION_ENTITY_ID,
            encryption_entity_type=ENCRYPTION_ENTITY_TYPE,
            encryption_source=ENCRYPTION_SOURCE,
            kind="dfareporting#conversionsBatchInsertRequest",
        )
        get_conn_mock.return_value.conversions.return_value.batchinsert.assert_called_once_with(
            profileId=PROFILE_ID, body=batch_request_mock.return_value
        )

        assert return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "campaign_manager.GoogleCampaignManagerHook.get_conn"
    )
    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks.campaign_manager.GoogleCampaignManagerHook"
        "._conversions_batch_request"
    )
    def test_conversions_batch_update(self, batch_request_mock, get_conn_mock):
        conversions = [{"conversions1": "value"}, {"conversions2": "value"}]

        return_value = {"hasFailures": False}
        get_conn_mock.return_value.conversions.return_value.batchupdate.return_value.execute.return_value = (
            return_value
        )

        batch_request_mock.return_value = "batch_request_mock"

        result = self.hook.conversions_batch_update(
            profile_id=PROFILE_ID,
            conversions=conversions,
            encryption_entity_id=ENCRYPTION_ENTITY_ID,
            encryption_entity_type=ENCRYPTION_ENTITY_TYPE,
            encryption_source=ENCRYPTION_SOURCE,
        )

        batch_request_mock.assert_called_once_with(
            conversions=conversions,
            encryption_entity_id=ENCRYPTION_ENTITY_ID,
            encryption_entity_type=ENCRYPTION_ENTITY_TYPE,
            encryption_source=ENCRYPTION_SOURCE,
            kind="dfareporting#conversionsBatchUpdateRequest",
        )
        get_conn_mock.return_value.conversions.return_value.batchupdate.assert_called_once_with(
            profileId=PROFILE_ID, body=batch_request_mock.return_value
        )

        assert return_value == result



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for Google BigQuery service.
"""

from __future__ import annotations

import os
from datetime import datetime

import pytest

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.bigquery import (
    BigQueryCreateEmptyDatasetOperator,
    BigQueryCreateTableOperator,
    BigQueryDeleteDatasetOperator,
)

try:
    from airflow.providers.google.cloud.transfers.bigquery_to_mysql import BigQueryToMySqlOperator
except ImportError:
    pytest.skip("MySQL not available", allow_module_level=True)

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
DAG_ID = "bigquery_to_mysql"

DATASET_NAME = f"dataset_{DAG_ID}_{ENV_ID}".replace("-", "_")
DATA_EXPORT_BUCKET_NAME = os.environ.get("GCP_BIGQUERY_EXPORT_BUCKET_NAME", "INVALID BUCKET NAME")
TABLE = "table_42"
destination_table = "mysql_table_test"
SCHEMA = [
    {"name": "emp_name", "type": "STRING", "mode": "REQUIRED"},
    {"name": "salary", "type": "INTEGER", "mode": "NULLABLE"},
]

with DAG(
    DAG_ID,
    schedule="@once",  # Override to match your needs
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "bigquery"],
) as dag:
    # [START howto_operator_bigquery_to_mysql]
    bigquery_to_mysql = BigQueryToMySqlOperator(
        task_id="bigquery_to_mysql",
        dataset_table=f"{DATASET_NAME}.{TABLE}",
        target_table_name=destination_table,
        replace=False,
    )
    # [END howto_operator_bigquery_to_mysql]

    create_dataset = BigQueryCreateEmptyDatasetOperator(task_id="create_dataset", dataset_id=DATASET_NAME)

    create_table = BigQueryCreateTableOperator(
        task_id="create_table",
        dataset_id=DATASET_NAME,
        table_id=TABLE,
        table_resource={
            "schema": {"fields": SCHEMA},
        },
    )

    delete_dataset = BigQueryDeleteDatasetOperator(
        task_id="delete_dataset", dataset_id=DATASET_NAME, delete_contents=True
    )

    (
        # TEST SETUP
        create_dataset
        >> create_table
        # TEST BODY
        >> bigquery_to_mysql
        # TEST TEARDOWN
        >> delete_dataset
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Storage operator."""

from __future__ import annotations

import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.exceptions import AirflowException, AirflowProviderDeprecationWarning
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.version_compat import BaseOperator

WILDCARD = "*"

if TYPE_CHECKING:
    from airflow.utils.context import Context


class GCSToGCSOperator(BaseOperator):
    """
    Copies objects from a bucket to another, with renaming if requested.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GCSToGCSOperator`

    :param source_bucket: The source Google Cloud Storage bucket where the
         object is. (templated)
    :param source_object: The source name of the object to copy in the Google cloud
        storage bucket. (templated)
        You can use only one wildcard for objects (filenames) within your
        bucket. The wildcard can appear inside the object name or at the
        end of the object name. Appending a wildcard to the bucket name is
        unsupported.
    :param source_objects: A list of source name of the objects to copy in the Google cloud
        storage bucket. (templated)
    :param destination_bucket: The destination Google Cloud Storage bucket
        where the object should be. If the destination_bucket is None, it defaults
        to source_bucket. (templated)
    :param destination_object: The destination name of the object in the
        destination Google Cloud Storage bucket. (templated)
        If a wildcard is supplied in the source_object argument, this is the
        prefix that will be prepended to the final destination objects' paths.
        Note that the source path's part before the wildcard will be removed;
        if it needs to be retained it should be appended to destination_object.
        For example, with prefix ``foo/*`` and destination_object ``blah/``, the
        file ``foo/baz`` will be copied to ``blah/baz``; to retain the prefix write
        the destination_object as e.g. ``blah/foo``, in which case the copied file
        will be named ``blah/foo/baz``.
        The same thing applies to source objects inside source_objects.
    :param move_object: When move object is True, the object is moved instead
        of copied to the new location. This is the equivalent of a mv command
        as opposed to a cp command.
    :param replace: Whether you want to replace existing destination files or not.
    :param delimiter: (Deprecated) This is used to restrict the result to only the 'files' in a given
        'folder'. If source_objects = ['foo/bah/'] and delimiter = '.avro', then only the 'files' in the
        folder 'foo/bah/' with '.avro' delimiter will be copied to the destination object.
    :param gcp_conn_id: (Optional) The connection ID used to connect to Google Cloud.
    :param last_modified_time: When specified, the objects will be copied or moved,
        only if they were modified after last_modified_time.
        If tzinfo has not been set, UTC will be assumed.
    :param maximum_modified_time: When specified, the objects will be copied or moved,
        only if they were modified before maximum_modified_time.
        If tzinfo has not been set, UTC will be assumed.
    :param is_older_than: When specified, the objects will be copied if they are older
        than the specified time in seconds.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param source_object_required: Whether you want to raise an exception when the source object
        doesn't exist. It doesn't have any effect when the source objects are folders or patterns.
    :param exact_match: When specified, only exact match of the source object (filename) will be
        copied.
    :param match_glob: (Optional) filters objects based on the glob pattern given by the string (
        e.g, ``'**/*/.json'``)

    :Example:

    The following Operator would copy a single file named
    ``sales/sales-2017/january.avro`` in the ``data`` bucket to the file named
    ``copied_sales/2017/january-backup.avro`` in the ``data_backup`` bucket ::

        copy_single_file = GCSToGCSOperator(
            task_id="copy_single_file",
            source_bucket="data",
            source_objects=["sales/sales-2017/january.avro"],
            destination_bucket="data_backup",
            destination_object="copied_sales/2017/january-backup.avro",
            exact_match=True,
            gcp_conn_id=google_cloud_conn_id,
        )

    The following Operator would copy all the Avro files from ``sales/sales-2017``
    folder (i.e. all files with names starting with that prefix) in ``data`` bucket to the
    ``copied_sales/2017`` folder in the ``data_backup`` bucket. ::

        copy_files = GCSToGCSOperator(
            task_id='copy_files',
            source_bucket='data',
            source_objects=['sales/sales-2017'],
            destination_bucket='data_backup',
            destination_object='copied_sales/2017/',
            match_glob='**/*.avro'
            gcp_conn_id=google_cloud_conn_id
        )

        Or ::

        copy_files = GCSToGCSOperator(
            task_id='copy_files',
            source_bucket='data',
            source_object='sales/sales-2017/*.avro',
            destination_bucket='data_backup',
            destination_object='copied_sales/2017/',
            gcp_conn_id=google_cloud_conn_id
        )

    The following Operator would move all the Avro files from ``sales/sales-2017``
    folder (i.e. all files with names starting with that prefix) in ``data`` bucket to the
    same folder in the ``data_backup`` bucket, deleting the original files in the
    process. ::

        move_files = GCSToGCSOperator(
            task_id="move_files",
            source_bucket="data",
            source_object="sales/sales-2017/*.avro",
            destination_bucket="data_backup",
            move_object=True,
            gcp_conn_id=google_cloud_conn_id,
        )

    The following Operator would move all the Avro files from ``sales/sales-2019``
     and ``sales/sales-2020`` folder in ``data`` bucket to the same folder in the
     ``data_backup`` bucket, deleting the original files in the process. ::

        move_files = GCSToGCSOperator(
            task_id="move_files",
            source_bucket="data",
            source_objects=["sales/sales-2019/*.avro", "sales/sales-2020"],
            destination_bucket="data_backup",
            delimiter=".avro",
            move_object=True,
            gcp_conn_id=google_cloud_conn_id,
        )

    """

    template_fields: Sequence[str] = (
        "source_bucket",
        "source_object",
        "source_objects",
        "destination_bucket",
        "destination_object",
        "delimiter",
        "impersonation_chain",
    )
    ui_color = "#f0eee4"

    def __init__(
        self,
        *,
        source_bucket,
        source_object=None,
        source_objects=None,
        destination_bucket=None,
        destination_object=None,
        delimiter=None,
        move_object=False,
        replace=True,
        gcp_conn_id="google_cloud_default",
        last_modified_time=None,
        maximum_modified_time=None,
        is_older_than=None,
        impersonation_chain: str | Sequence[str] | None = None,
        source_object_required=False,
        exact_match=False,
        match_glob: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.source_bucket = source_bucket
        if source_object and WILDCARD in source_object:
            warnings.warn(
                "Usage of wildcard (*) in 'source_object' is deprecated, utilize 'match_glob' instead",
                AirflowProviderDeprecationWarning,
                stacklevel=2,
            )
        self.source_object = source_object
        if source_objects and any(WILDCARD in obj for obj in source_objects):
            warnings.warn(
                "Usage of wildcard (*) in 'source_objects' is deprecated, utilize 'match_glob' instead",
                AirflowProviderDeprecationWarning,
                stacklevel=2,
            )
        self.source_objects = source_objects
        self.destination_bucket = destination_bucket
        self.destination_object = destination_object
        if delimiter:
            warnings.warn(
                "Usage of 'delimiter' is deprecated, please use 'match_glob' instead",
                AirflowProviderDeprecationWarning,
                stacklevel=2,
            )
        self.delimiter = delimiter
        self.move_object = move_object
        self.replace = replace
        self.gcp_conn_id = gcp_conn_id
        self.last_modified_time = last_modified_time
        self.maximum_modified_time = maximum_modified_time
        self.is_older_than = is_older_than
        self.impersonation_chain = impersonation_chain
        self.source_object_required = source_object_required
        self.exact_match = exact_match
        self.match_glob = match_glob

    def execute(self, context: Context):
        hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        if self.source_objects and self.source_object:
            error_msg = (
                f"You can either set source_object parameter or source_objects parameter but not both. "
                f"Found source_object={self.source_object} and source_objects={self.source_objects}"
            )
            raise AirflowException(error_msg)

        if not self.source_object and not self.source_objects:
            error_msg = "You must set source_object parameter or source_objects parameter. None set"
            raise AirflowException(error_msg)

        if self.source_objects and not all(isinstance(item, str) for item in self.source_objects):
            raise AirflowException("At least, one of the `objects` in the `source_objects` is not a string")

        # If source_object is set, default it to source_objects
        if self.source_object:
            self.source_objects = [self.source_object]

        if self.destination_bucket is None:
            self.log.warning(
                "destination_bucket is None. Defaulting it to source_bucket (%s)", self.source_bucket
            )
            self.destination_bucket = self.source_bucket

        # An empty source_object means to copy all files
        if len(self.source_objects) == 0:
            self.source_objects = [""]
        # Raise exception if empty string `''` is used twice in source_object, this is to avoid double copy
        if self.source_objects.count("") > 1:
            raise AirflowException("You can't have two empty strings inside source_object")

        # Iterate over the source_objects and do the copy
        for prefix in self.source_objects:
            # Check if prefix contains wildcard
            if WILDCARD in prefix:
                self._copy_source_with_wildcard(hook=hook, prefix=prefix)
            # Now search with prefix using provided delimiter if any
            else:
                self._copy_source_without_wildcard(hook=hook, prefix=prefix)

    def _ignore_existing_files(self, hook, prefix, **kwargs):
        # list all files in the Destination GCS bucket
        # and only keep those files which are present in
        # Source GCS bucket and not in Destination GCS bucket
        delimiter = kwargs.get("delimiter")
        match_glob = kwargs.get("match_glob")
        objects = kwargs.get("objects")
        if self.destination_object is None:
            existing_objects = hook.list(
                self.destination_bucket, prefix=prefix, delimiter=delimiter, match_glob=match_glob
            )
        else:
            self.log.info("Replaced destination_object with source_object prefix.")
            destination_objects = hook.list(
                self.destination_bucket,
                prefix=self.destination_object,
                delimiter=delimiter,
                match_glob=match_glob,
            )
            existing_objects = [
                dest_object.replace(self.destination_object, prefix, 1) for dest_object in destination_objects
            ]

        objects = list(set(objects) - set(existing_objects))
        if objects:
            self.log.info("%s files are going to be synced: %s.", len(objects), objects)
        else:
            self.log.info("There are no new files to sync. Have a nice day!")
        return objects

    def _copy_source_without_wildcard(self, hook, prefix):
        """
        List all files in source_objects, copy files to destination_object, and rename each source file.

        For source_objects with no wildcard, this operator would first list
        all files in source_objects, using provided delimiter if any. Then copy
        files from source_objects to destination_object and rename each source
        file. Note that if the flag exact_match=False, then each item in the source_objects
        (or source_object itself) will be considered as a prefix for the source objects search.

        Example 1:


        The following Operator would copy all the files from ``a/`` folder
        (i.e a/a.csv, a/b.csv, a/c.csv) in ``data`` bucket to the ``b/`` folder in
        the ``data_backup`` bucket (b/a.csv, b/b.csv, b/c.csv) ::

            copy_files = GCSToGCSOperator(
                task_id="copy_files_without_wildcard",
                source_bucket="data",
                source_objects=["a/"],
                destination_bucket="data_backup",
                destination_object="b/",
                gcp_conn_id=google_cloud_conn_id,
            )

        Example 2:


        The following Operator would copy all avro files from ``a/`` folder
        (i.e a/a.avro, a/b.avro, a/c.avro) in ``data`` bucket to the ``b/`` folder in
        the ``data_backup`` bucket (b/a.avro, b/b.avro, b/c.avro) ::

            copy_files = GCSToGCSOperator(
                task_id="copy_files_without_wildcard",
                source_bucket="data",
                source_objects=["a/"],
                destination_bucket="data_backup",
                destination_object="b/",
                delimiter=".avro",
                gcp_conn_id=google_cloud_conn_id,
            )

        Example 3:


        The following Operator would copy files (a/file_1.txt, a/file_2.csv, a/file_3.avro)
        in ``data`` bucket to the ``b/`` folder in
        the ``data_backup`` bucket (b/file_1.txt, b/file_2.csv, b/file_3.avro) ::

            copy_files = GCSToGCSOperator(
                task_id="copy_files_without_wildcard",
                source_bucket="data",
                source_objects=["a/file_1.txt", "a/file_2.csv", "a/file_3.avro"],
                destination_bucket="data_backup",
                destination_object="b/",
                gcp_conn_id=google_cloud_conn_id,
            )

        Example 4:

        The following Operator would copy files corresponding to the prefix 'a/foo.txt'
        (a/foo.txt, a/foo.txt.abc, a/foo.txt/subfolder/file.txt) in ``data`` bucket to
        the ``b/`` folder in the ``data_backup`` bucket
        (b/foo.txt, b/foo.txt.abc, b/foo.txt/subfolder/file.txt) ::

            copy_files = GCSToGCSOperator(
                task_id="copy_files_without_wildcard",
                source_bucket="data",
                source_object="a/foo.txt",
                destination_bucket="data_backup",
                destination_object="b/",
                gcp_conn_id=google_cloud_conn_id,
            )
        """
        objects = hook.list(
            self.source_bucket, prefix=prefix, delimiter=self.delimiter, match_glob=self.match_glob
        )

        objects = [obj for obj in objects if self._check_exact_match(obj, prefix)]

        if not self.replace:
            # If we are not replacing, ignore files already existing in source buckets
            objects = self._ignore_existing_files(
                hook, prefix, objects=objects, delimiter=self.delimiter, match_glob=self.match_glob
            )

        # If objects is empty, and we have prefix, let's check if prefix is a blob
        # and copy directly
        if len(objects) == 0 and prefix:
            if hook.exists(self.source_bucket, prefix):
                self._copy_single_object(
                    hook=hook, source_object=prefix, destination_object=self.destination_object
                )
            elif self.source_object_required:
                msg = f"{prefix} does not exist in bucket {self.source_bucket}"
                self.log.warning(msg)
                raise AirflowException(msg)
        if len(objects) == 1 and objects[0][-1] != "/":
            self._copy_file(hook=hook, source_object=objects[0])
        elif len(objects):
            self._copy_multiple_objects(hook=hook, source_objects=objects, prefix=prefix)

    def _copy_file(self, hook, source_object):
        destination_object = self.destination_object or source_object
        if self.destination_object and self.destination_object[-1] == "/":
            file_name = source_object.split("/")[-1]
            destination_object += file_name
        self._copy_single_object(
            hook=hook, source_object=source_object, destination_object=destination_object
        )

    def _copy_multiple_objects(self, hook, source_objects, prefix):
        # Check whether the prefix is a root directory for all the rest of objects.
        _pref = prefix.rstrip("/")
        is_directory = prefix.endswith("/") or all(
            obj.replace(_pref, "", 1).startswith("/") for obj in source_objects
        )

        if is_directory:
            base_path = prefix.rstrip("/") + "/"
        else:
            base_path = prefix[0 : prefix.rfind("/") + 1] if "/" in prefix else ""

        for source_obj in source_objects:
            if not self._check_exact_match(source_obj, prefix):
                continue
            if self.destination_object is None:
                destination_object = source_obj
            else:
                file_name_postfix = source_obj.replace(base_path, "", 1)
                destination_object = self.destination_object.rstrip("/") + "/" + file_name_postfix

            self._copy_single_object(
                hook=hook, source_object=source_obj, destination_object=destination_object
            )

    def _check_exact_match(self, source_object: str, prefix: str) -> bool:
        """Check whether source_object's name matches the prefix according to the exact_match flag."""
        if self.exact_match and (source_object != prefix or not source_object.endswith(prefix)):
            return False
        return True

    def _copy_source_with_wildcard(self, hook, prefix):
        total_wildcards = prefix.count(WILDCARD)
        if total_wildcards > 1:
            error_msg = (
                "Only one wildcard '*' is allowed in source_object parameter. "
                f"Found {total_wildcards} in {prefix}."
            )

            raise AirflowException(error_msg)
        self.log.info("Delimiter ignored because wildcard is in prefix")
        prefix_, delimiter = prefix.split(WILDCARD, 1)
        objects = hook.list(self.source_bucket, prefix=prefix_, delimiter=delimiter)
        # TODO: After deprecating delimiter and wildcards in source objects,
        #       remove previous line and uncomment the following:
        # match_glob = f"**/*{delimiter}" if delimiter else None
        # objects = hook.list(self.source_bucket, prefix=prefix_, match_glob=match_glob)
        if not self.replace:
            # If we are not replacing, list all files in the Destination GCS bucket
            # and only keep those files which are present in
            # Source GCS bucket and not in Destination GCS bucket
            objects = self._ignore_existing_files(hook, prefix_, delimiter=delimiter, objects=objects)
            # TODO: After deprecating delimiter and wildcards in source objects,
            #       remove previous line and uncomment the following:
            # objects = self._ignore_existing_files(hook, prefix_, match_glob=match_glob, objects=objects)

        for source_object in objects:
            if self.destination_object is None:
                destination_object = source_object
            else:
                destination_object = source_object.replace(prefix_, self.destination_object, 1)

            self._copy_single_object(
                hook=hook, source_object=source_object, destination_object=destination_object
            )

    def _copy_single_object(self, hook, source_object, destination_object):
        if self.is_older_than:
            # Here we check if the given object is older than the given time
            # If given, last_modified_time and maximum_modified_time is ignored
            if hook.is_older_than(self.source_bucket, source_object, self.is_older_than):
                self.log.info("Object is older than %s seconds ago", self.is_older_than)
            else:
                self.log.debug("Object is not older than %s seconds ago", self.is_older_than)
                return
        elif self.last_modified_time and self.maximum_modified_time:
            # check to see if object was modified between last_modified_time and
            # maximum_modified_time
            if hook.is_updated_between(
                self.source_bucket, source_object, self.last_modified_time, self.maximum_modified_time
            ):
                self.log.info(
                    "Object has been modified between %s and %s",
                    self.last_modified_time,
                    self.maximum_modified_time,
                )
            else:
                self.log.debug(
                    "Object was not modified between %s and %s",
                    self.last_modified_time,
                    self.maximum_modified_time,
                )
                return

        elif self.last_modified_time is not None:
            # Check to see if object was modified after last_modified_time
            if hook.is_updated_after(self.source_bucket, source_object, self.last_modified_time):
                self.log.info("Object has been modified after %s ", self.last_modified_time)
            else:
                self.log.debug("Object was not modified after %s ", self.last_modified_time)
                return
        elif self.maximum_modified_time is not None:
            # Check to see if object was modified before maximum_modified_time
            if hook.is_updated_before(self.source_bucket, source_object, self.maximum_modified_time):
                self.log.info("Object has been modified before %s ", self.maximum_modified_time)
            else:
                self.log.debug("Object was not modified before %s ", self.maximum_modified_time)
                return

        self.log.info(
            "Executing copy of gs://%s/%s to gs://%s/%s",
            self.source_bucket,
            source_object,
            self.destination_bucket,
            destination_object,
        )
        hook.rewrite(self.source_bucket, source_object, self.destination_bucket, destination_object)

        if self.move_object:
            hook.delete(self.source_bucket, source_object)

    def get_openlineage_facets_on_complete(self, task_instance):
        """
        Implement _on_complete because execute method does preprocessing on internals.

        This means we won't have to normalize self.source_object and self.source_objects,
        destination bucket and so on.
        """
        from airflow.providers.common.compat.openlineage.facet import Dataset
        from airflow.providers.google.cloud.openlineage.utils import extract_ds_name_from_gcs_path
        from airflow.providers.openlineage.extractors import OperatorLineage

        inputs = [extract_ds_name_from_gcs_path(path) for path in self.source_objects]

        if self.destination_object is None:
            outputs = inputs.copy()
        else:
            outputs = [extract_ds_name_from_gcs_path(self.destination_object)]

        return OperatorLineage(
            inputs=[
                Dataset(namespace=f"gs://{self.source_bucket}", name=source) for source in sorted(set(inputs))
            ],
            outputs=[
                Dataset(namespace=f"gs://{self.destination_bucket}", name=target)
                for target in sorted(set(outputs))
            ],
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Vertex AI hook."""

from __future__ import annotations

import dataclasses
from collections.abc import MutableMapping
from typing import Any

import vertex_ray
from google.cloud import aiplatform
from google.cloud.aiplatform.vertex_ray.util import resources
from google.cloud.aiplatform_v1 import (
    PersistentResourceServiceClient,
)
from proto.marshal.collections.repeated import Repeated

from airflow.providers.google.common.hooks.base_google import GoogleBaseHook


class RayHook(GoogleBaseHook):
    """Hook for Google Cloud Vertex AI Ray APIs."""

    def extract_cluster_id(self, cluster_path) -> str:
        """Extract cluster_id from cluster_path."""
        cluster_id = PersistentResourceServiceClient.parse_persistent_resource_path(cluster_path)[
            "persistent_resource"
        ]
        return cluster_id

    def serialize_cluster_obj(self, cluster_obj: resources.Cluster) -> dict:
        """Serialize Cluster dataclass to dict."""

        def __encode_value(value: Any) -> Any:
            if isinstance(value, (list, Repeated)):
                return [__encode_value(nested_value) for nested_value in value]
            if not isinstance(value, dict) and isinstance(value, MutableMapping):
                return {key: __encode_value(nested_value) for key, nested_value in dict(value).items()}
            if dataclasses.is_dataclass(value):
                return dataclasses.asdict(value)
            return value

        return {
            field.name: __encode_value(getattr(cluster_obj, field.name))
            for field in dataclasses.fields(cluster_obj)
        }

    @GoogleBaseHook.fallback_to_default_project_id
    def create_ray_cluster(
        self,
        project_id: str,
        location: str,
        head_node_type: resources.Resources = resources.Resources(),
        python_version: str = "3.10",
        ray_version: str = "2.33",
        network: str | None = None,
        service_account: str | None = None,
        cluster_name: str | None = None,
        worker_node_types: list[resources.Resources] | None = None,
        custom_images: resources.NodeImages | None = None,
        enable_metrics_collection: bool = True,
        enable_logging: bool = True,
        psc_interface_config: resources.PscIConfig | None = None,
        reserved_ip_ranges: list[str] | None = None,
        labels: dict[str, str] | None = None,
    ) -> str:
        """
        Create a Ray cluster on the Vertex AI.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param head_node_type: The head node resource. Resources.node_count must be 1. If not set, default
            value of Resources() class will be used.
        :param python_version: Python version for the ray cluster.
        :param ray_version: Ray version for the ray cluster. Default is 2.33.0.
        :param network: Virtual private cloud (VPC) network. For Ray Client, VPC peering is required to
            connect to the Ray Cluster managed in the Vertex API service. For Ray Job API, VPC network is not
            required because Ray Cluster connection can be accessed through dashboard address.
        :param service_account: Service account to be used for running Ray programs on the cluster.
        :param cluster_name: This value may be up to 63 characters, and valid characters are `[a-z0-9_-]`.
            The first character cannot be a number or hyphen.
        :param worker_node_types: The list of Resources of the worker nodes. The same Resources object should
            not appear multiple times in the list.
        :param custom_images: The NodeImages which specifies head node and worker nodes images. All the
            workers will share the same image. If each Resource has a specific custom image, use
            `Resources.custom_image` for head/worker_node_type(s). Note that configuring
            `Resources.custom_image` will override `custom_images` here. Allowlist only.
        :param enable_metrics_collection: Enable Ray metrics collection for visualization.
        :param enable_logging: Enable exporting Ray logs to Cloud Logging.
        :param psc_interface_config: PSC-I config.
        :param reserved_ip_ranges: A list of names for the reserved IP ranges under the VPC network that can
            be used for this cluster. If set, we will deploy the cluster within the provided IP ranges.
            Otherwise, the cluster is deployed to any IP ranges under the provided VPC network.
            Example: ["vertex-ai-ip-range"].
        :param labels: The labels with user-defined metadata to organize Ray cluster.
            Label keys and values can be no longer than 64 characters (Unicode codepoints), can only contain
            lowercase letters, numeric characters, underscores and dashes. International characters are allowed.
            See https://goo.gl/xmQnxf for more information and examples of labels.
        """
        aiplatform.init(project=project_id, location=location, credentials=self.get_credentials())
        cluster_path = vertex_ray.create_ray_cluster(
            head_node_type=head_node_type,
            python_version=python_version,
            ray_version=ray_version,
            network=network,
            service_account=service_account,
            cluster_name=cluster_name,
            worker_node_types=worker_node_types,
            custom_images=custom_images,
            enable_metrics_collection=enable_metrics_collection,
            enable_logging=enable_logging,
            psc_interface_config=psc_interface_config,
            reserved_ip_ranges=reserved_ip_ranges,
            labels=labels,
        )
        return cluster_path

    @GoogleBaseHook.fallback_to_default_project_id
    def list_ray_clusters(
        self,
        project_id: str,
        location: str,
    ) -> list[resources.Cluster]:
        """
        List Ray clusters under the currently authenticated project.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        """
        aiplatform.init(project=project_id, location=location, credentials=self.get_credentials())
        ray_clusters = vertex_ray.list_ray_clusters()
        return ray_clusters

    @GoogleBaseHook.fallback_to_default_project_id
    def get_ray_cluster(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
    ) -> resources.Cluster:
        """
        Get Ray cluster.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param cluster_id: Cluster resource ID.
        """
        aiplatform.init(project=project_id, location=location, credentials=self.get_credentials())
        ray_cluster_name = PersistentResourceServiceClient.persistent_resource_path(
            project=project_id,
            location=location,
            persistent_resource=cluster_id,
        )
        ray_cluster = vertex_ray.get_ray_cluster(
            cluster_resource_name=ray_cluster_name,
        )
        return ray_cluster

    @GoogleBaseHook.fallback_to_default_project_id
    def update_ray_cluster(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        worker_node_types: list[resources.Resources],
    ) -> str:
        """
        Update Ray cluster (currently support resizing node counts for worker nodes).

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param cluster_id: Cluster resource ID.
        :param worker_node_types: The list of Resources of the resized worker nodes. The same Resources
            object should not appear multiple times in the list.
        """
        aiplatform.init(project=project_id, location=location, credentials=self.get_credentials())
        ray_cluster_name = PersistentResourceServiceClient.persistent_resource_path(
            project=project_id,
            location=location,
            persistent_resource=cluster_id,
        )
        updated_ray_cluster_name = vertex_ray.update_ray_cluster(
            cluster_resource_name=ray_cluster_name, worker_node_types=worker_node_types
        )
        return updated_ray_cluster_name

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_ray_cluster(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
    ) -> None:
        """
        Delete Ray cluster.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param cluster_id: Cluster resource ID.
        """
        aiplatform.init(project=project_id, location=location, credentials=self.get_credentials())
        ray_cluster_name = PersistentResourceServiceClient.persistent_resource_path(
            project=project_id,
            location=location,
            persistent_resource=cluster_id,
        )
        vertex_ray.delete_ray_cluster(cluster_resource_name=ray_cluster_name)
